
import io
import logging
import os
from pathlib import Path
import subprocess
import sys

import asyncio
from dan.core.terminal import write as term_write


class CommandError(RuntimeError):
    def __init__(self, message, rc, stdout, stderr) -> None:
        super().__init__(message)
        self.rc = rc
        self.stdout = stdout
        self.stderr = stderr


_encoding = 'cp1252' if os.name == 'nt' else 'utf-8'


async def log_stream(stream, *files):
    with stream as lines:
        async for line in lines:
            for file in files:
                if file in [sys.stdout, sys.stderr]:
                    term_write(line, end='') #, file=file)
                else:
                    file.write(line)


class AsyncStreamProducer:
    
    class Iterator:
        def __init__(self) -> None:
            self.q = asyncio.Queue()

        async def __aiter__(self):
            while True:
                try:
                    line = await self.q.get()
                except asyncio.CancelledError:
                    return
                if line is None:
                    self.q.task_done()
                    return
                yield line
                self.q.task_done()

    def __init__(self, stream) -> None:
        self.stream = stream
        self.__iterators: list[AsyncStreamProducer.Iterator] = list()

    async def __notify(self, data):
        asyncio.gather(*[it.q.put(data) for it in self.__iterators])
        asyncio.gather(*[it.q.join() for it in self.__iterators])
    
    async def consume(self):
        while not self.stream.at_eof():
            data = await self.stream.readline()
            if len(data) == 0:
                continue
            line = data.decode(_encoding)
            await self.__notify(line)

        await self.__notify(None)
    
    def __enter__(self):
        it = self.Iterator()
        self.__iterators.append(it)
        return it

    def __exit__(self, *exc):
        pass


_jobs_sem: asyncio.Semaphore = None


def max_jobs(count=1):
    global _jobs_sem
    if count > 0:
        _jobs_sem = asyncio.Semaphore(count)
    else:
        _jobs_sem = None

def cmdline2list(s: str):
    """
    Translate a command line string into a sequence of arguments,
    using the same rules as the MS C runtime:

    1) Arguments are delimited by white space, which is either a
       space or a tab.

    2) A string surrounded by double quotation marks is
       interpreted as a single argument, regardless of white space
       contained within.  A quoted string can be embedded in an
       argument.

    3) A double quotation mark preceded by a backslash is
       interpreted as a literal double quotation mark.

    4) Backslashes are interpreted literally, unless they
       immediately precede a double quotation mark.

    5) If backslashes immediately precede a double quotation mark,
       every pair of backslashes is interpreted as a literal
       backslash.  If the number of backslashes is odd, the last
       backslash escapes the next double quotation mark as
       described in rule 3.
    """
    result = []
    current = []
    quote = None
    escaped = False
    for c in s:
        match c:
            case ' ' | '\t':
                if escaped:
                    current.append(c)
                    escaped = False
                else:
                    if quote is None:
                        if current:
                            result.append(''.join(current))
                            current = []
                    else:
                        current.append(c)
            case "'" | '"':
                if escaped:
                    current.append(c)
                    escaped = False
                else:                    
                    if quote is not None and quote == c:
                        quote = None
                    else:
                        quote = c
            case '\\':
                if escaped:
                    current.append('\\')
                escaped = True
            case _:
                if escaped:
                    current.append('\\')
                current.append(c)
                escaped = False

    if current:
        result.append(''.join(current))

    return result

def list2cmdline(command: list|str):
    if isinstance(command, list):
        cmd = list()
        for part in command:
            if isinstance(part, Path):
                cmd.append(part.as_posix())
            else:
                cmd.append(part)
        return subprocess.list2cmdline(cmd)
    return command


async def async_run(command, log=True, logger: logging.Logger = None, no_raise=False, env=None, cwd=None, out_capture=None, err_capture=None, all_capture=None, input: str = None) -> tuple[str, str, int]:
    if _jobs_sem is not None:
        await _jobs_sem.acquire()
    try:
        command = list2cmdline(command)
        if env is not None:
            e = dict(os.environ)
            for k, v in env.items():
                e[k] = v
            env = e
        if input is not None:
            stdin = asyncio.subprocess.PIPE
        else:
            stdin = None
        if logger is not None:
            logger.debug('executing: %s', command)
        proc = await asyncio.subprocess.create_subprocess_shell(command,
                                                                stdout=asyncio.subprocess.PIPE,
                                                                stderr=asyncio.subprocess.PIPE,
                                                                stdin=stdin,
                                                                env=env,
                                                                cwd=cwd)
        out = io.StringIO()
        err = io.StringIO()
        outs = [out]
        errs = [err]
        if log:
            outs.append(sys.stdout)
            errs.append(sys.stderr)

        out_iter = AsyncStreamProducer(proc.stdout)
        err_iter = AsyncStreamProducer(proc.stderr)

        futures = [
            log_stream(out_iter, *outs),
            log_stream(err_iter, *errs),
        ]

        if all_capture is not None:
            out_capture = err_capture = all_capture

        if out_capture is not None:
            futures.append(out_capture(out_iter))

        if err_capture is not None:
            futures.append(out_capture(err_iter))

        futures.extend([
            out_iter.consume(),
            err_iter.consume(),
        ])
        
        if input is not None:
            proc.stdin.write(input.encode())
            proc.stdin.write_eof()
    
        await asyncio.gather(
            *futures,
            proc.wait())
        # make sure return code is available
        await proc.communicate()
        out = out.getvalue()
        err = err.getvalue()
        if proc.returncode != 0 and not no_raise:
            message = f'command returned {proc.returncode}: {command}\n{out}\n{err}'
            if logger:
                logger.error(message)
            raise CommandError(message, proc.returncode, out, err)
        return out, err, proc.returncode
    finally:
        if _jobs_sem is not None:
            _jobs_sem.release()


def sync_run(command, pipe=True, logger: logging.Logger = None, no_raise=False, shell=True, env=None, cwd=None):
    command = list2cmdline(command)
    if pipe:
        stdout = subprocess.PIPE
    else:
        stdout = None
    if logger:
        logger.debug(f'executing: {command}')
    if env:
        e = dict(os.environ)
        for k, v in env.items():
            e[k] = v
        env = e
    proc = subprocess.Popen(command,
                            stdout=stdout,
                            stderr=stdout,
                            shell=shell,
                            env=env,
                            cwd=cwd,
                            universal_newlines=True)
    out, err = proc.communicate()
    if proc.returncode != 0 and not no_raise:
        message = f'command returned {proc.returncode}: {command}\n{err if err else ""}'
        if logger:
            logger.error(message)
        raise CommandError(message, proc.returncode, out, err)
    return out if out else None, err if err else None, proc.returncode



from asyncio import *
from async_property import *

import threading
import concurrent.futures
import multiprocessing
import typing as t
import inspect
import threading

from dan.core.functools import BaseDecorator


async def may_await(obj):
    """Awaits given object if it is an awaitable

    If obj is an awaitable it's await-result is returned.
    If obj is not an awaitable it is returned directly.
    """
    if inspect.isawaitable(obj):
        return await obj
    else:
        return obj


class Cached(BaseDecorator):
    def __init__(self, fn, unique=False):
        self.__fn = fn
        self.__unique = unique
        self.__is_method = fn.__call__.__class__.__name__ == "method-wrapper"
        if not self.__is_method and self.__unique:
            self.__cache = None
        else:
            self.__cache: dict[int, Future] = dict()

    async def __call__(self, *args, **kwds):
        if not self.__is_method and self.__unique:
            if self.__cache is None:
                self.__cache = Future()
                try:
                    self.__cache.set_result(await self.__fn(*args, **kwds))
                except Exception as ex:
                    self.__cache.set_exception(ex)
            elif not self.__cache.done():
                await self.__cache
            return self.__cache.result()
        else:
            key = id(args[0]) if self.__unique else hash((args, frozenset(kwds)))
            if key not in self.__cache:
                self.__cache[key] = Future()
                try:
                    self.__cache[key].set_result(await self.__fn(*args, **kwds))
                except Exception as ex:
                    self.__cache[key].set_exception(ex)
            elif not self.__cache[key].done():
                await self.__cache[key]

            return self.__cache[key].result()

    def clear_all(self):
        if self.__unique:
            self.__cache = dict()
        else:
            self.__cache = None


def cached(*args, **kwargs):
    if len(args) == 1 and callable(args[0]):
        return Cached(args[0])
    else:

        def wrapper(fn):
            return Cached(fn, *args, **kwargs)

        return wrapper


class _SyncWaitThread(threading.Thread):
    def __init__(self, coro):
        self.coro = coro
        self.result = None
        self.err = None
        super().__init__()

    def run(self):
        try:
            loop = new_event_loop()
            self.result = loop.run_until_complete(self.coro)
            tasks = all_tasks(loop)
            if len(tasks):
                for t in tasks:
                    t.cancel()
                loop.run_until_complete(wait(tasks))
        except Exception as err:
            self.err = err


def sync_wait(coro):
    thread = _SyncWaitThread(coro)
    thread.start()
    thread.join()
    if thread.err:
        raise thread.err
    return thread.result


_async_pool = concurrent.futures.ThreadPoolExecutor(
    max_workers=multiprocessing.cpu_count(),
)


async def async_wait(fn, *args, **kwargs):
    loop = get_running_loop()

    def wrapper():
        return fn(*args, **kwargs)

    return await loop.run_in_executor(_async_pool, wrapper)


class ExceptionGroup(Exception):
    def __init__(self, message: str, errors: t.Iterable[Exception] = set()) -> None:
        super().__init__(message)
        self.errors: set[Exception] = set(errors)

    def add(self, e: Exception):
        self.errors.add(e)

    def __str__(self) -> str:
        s = super().__str__()
        for e in self.errors:
            s += "\n" + str(e)
        return s


class TaskGroup:
    """Asynchronous context manager for managing groups of tasks.

    Example use:

        async with asyncio.TaskGroup('my group name') as group:
            task1 = group.create_task(some_coroutine(...))
            task2 = group.create_task(other_coroutine(...))
        print("Both tasks have completed now.")

    All tasks are awaited when the context manager exits.

    Any exceptions other than `asyncio.CancelledError` raised within
    a task will cancel all remaining tasks and wait for them to exit.
    The exceptions are then combined and raised as an `ExceptionGroup`.
    """

    def __init__(self, name="a TaskGroup"):
        self._entered = False
        self._exiting = False
        self._aborting = False
        self._loop = None
        self._parent_task = None
        self._parent_cancel_requested = False
        self._tasks: set[Task] = set()
        self._errors = []
        self._results = []
        self._base_error = None
        self._on_completed_fut = None
        self._name = name

    def __repr__(self):
        info = [""]
        if self._tasks:
            info.append(f"tasks={len(self._tasks)}")
        if self._errors:
            info.append(f"errors={len(self._errors)}")
        if self._aborting:
            info.append("cancelling")
        elif self._entered:
            info.append("entered")

        info_str = " ".join(info)
        return f"<TaskGroup{info_str}>"

    async def __aenter__(self):
        if self._entered:
            raise RuntimeError(f"TaskGroup {self!r} has been already entered")
        self._entered = True

        if self._loop is None:
            self._loop = get_running_loop()

        self._parent_task = current_task(self._loop)
        if self._parent_task is None:
            raise RuntimeError(f"TaskGroup {self!r} cannot determine the parent task")

        return self

    async def __aexit__(self, et, exc, tb):
        self._exiting = True

        if exc is not None and self._is_base_error(exc) and self._base_error is None:
            self._base_error = exc

        propagate_cancellation_error = exc if et is CancelledError else None
        if self._parent_cancel_requested:
            # If this flag is set we *must* call uncancel().
            if self._parent_task.uncancel() == 0:
                # If there are no pending cancellations left,
                # don't propagate CancelledError.
                propagate_cancellation_error = None

        if et is not None:
            if not self._aborting:
                # Our parent task is being cancelled:
                #
                #    async with TaskGroup() as g:
                #        g.create_task(...)
                #        await ...  # <- CancelledError
                #
                # or there's an exception in "async with":
                #
                #    async with TaskGroup() as g:
                #        g.create_task(...)
                #        1 / 0
                #
                self._abort()

        # We use while-loop here because "self._on_completed_fut"
        # can be cancelled multiple times if our parent task
        # is being cancelled repeatedly (or even once, when
        # our own cancellation is already in progress)
        while self._tasks:
            if self._on_completed_fut is None:
                self._on_completed_fut = self._loop.create_future()
            try:
                await self._on_completed_fut
            except CancelledError as ex:
                if not self._aborting:
                    # Our parent task is being cancelled:
                    #
                    #    async def wrapper():
                    #        async with TaskGroup() as g:
                    #            g.create_task(foo)
                    #
                    # "wrapper" is being cancelled while "foo" is
                    # still running.
                    propagate_cancellation_error = ex
                    self._abort()

            self._on_completed_fut = None

        assert not self._tasks

        if self._base_error is not None:
            raise self._base_error

        # Propagate CancelledError if there is one, except if there
        # are other errors -- those have priority.
        if propagate_cancellation_error and not self._errors:
            raise propagate_cancellation_error

        if et is not None and et is not CancelledError:
            self._errors.append(exc)

        if self._errors:
            # Exceptions are heavy objects that can have object
            # cycles (bad for GC); let's not keep a reference to
            # a bunch of them.
            try:
                me = ExceptionGroup(f"unhandled errors in {self._name}", self._errors)
                raise me from None
            finally:
                self._errors = None

    def create_task(self, coro, *, name=None, context=None):
        """Create a new task in this group and return it.

        Similar to `asyncio.create_task`.
        """
        if not self._entered:
            raise RuntimeError(f"TaskGroup {self!r} has not been entered")
        if self._exiting and not self._tasks:
            raise RuntimeError(f"TaskGroup {self!r} is finished")
        if self._aborting:
            raise RuntimeError(f"TaskGroup {self!r} is shutting down")
        if isfuture(coro):
            task = coro
        elif context is None:
            task = self._loop.create_task(coro, name=name)
        else:
            task = self._loop.create_task(coro, context=context, name=name)
        task.add_done_callback(self._on_task_done)
        self._tasks.add(task)
        return task

    # Since Python 3.8 Tasks propagate all exceptions correctly,
    # except for KeyboardInterrupt and SystemExit which are
    # still considered special.

    def _is_base_error(self, exc: BaseException) -> bool:
        assert isinstance(exc, BaseException)
        return isinstance(exc, (SystemExit, KeyboardInterrupt))

    def _abort(self):
        self._aborting = True

        for t in self._tasks:
            if not t.done():
                t.cancel()

    def _on_task_done(self, task):
        self._tasks.discard(task)

        if self._on_completed_fut is not None and not self._tasks:
            if not self._on_completed_fut.done():
                self._on_completed_fut.set_result(True)

        if task.cancelled():
            return

        exc = task.exception()
        if exc is None:
            self._results.append(task.result())
            return

        self._errors.append(exc)
        if self._is_base_error(exc) and self._base_error is None:
            self._base_error = exc

        if self._parent_task.done():
            # Not sure if this case is possible, but we want to handle
            # it anyways.
            self._loop.call_exception_handler(
                {
                    "message": f"Task {task!r} has errored out but its parent "
                    f"task {self._parent_task} is already completed",
                    "exception": exc,
                    "task": task,
                }
            )
            return

        if not self._aborting and not self._parent_cancel_requested:
            # If parent task *is not* being cancelled, it means that we want
            # to manually cancel it to abort whatever is being run right now
            # in the TaskGroup.  But we want to mark parent task as
            # "not cancelled" later in __aexit__.  Example situation that
            # we need to handle:
            #
            #    async def foo():
            #        try:
            #            async with TaskGroup() as g:
            #                g.create_task(crash_soon())
            #                await something  # <- this needs to be canceled
            #                                 #    by the TaskGroup, e.g.
            #                                 #    foo() needs to be cancelled
            #        except Exception:
            #            # Ignore any exceptions raised in the TaskGroup
            #            pass
            #        await something_else     # this line has to be called
            #                                 # after TaskGroup is finished.
            self._abort()
            self._parent_cancel_requested = True
            self._parent_task.cancel()

    def results(self):
        if not self._entered:
            raise RuntimeError(f"TaskGroup {self!r} has not been entered")
        if self._exiting and self._tasks:
            raise RuntimeError(f"TaskGroup {self!r} is not finished")
        return self._results


class async_lock:
    def __init__(self, lock):
        self.lock = lock

    async def __aenter__(self):
        loop = get_event_loop()
        await loop.run_in_executor(_async_pool, self.lock.acquire)

    async def __aexit__(self, *args):
        self.lock.release()


class ThreadLock:
    def __init__(self) -> None:
        self._lk = threading.Lock()

    def __enter__(self):
        self._lk.acquire()

    def __exit__(self, *exc):
        self._lk.release()

    async def __aenter__(self):
        loop = get_event_loop()
        await loop.run_in_executor(_async_pool, self._lk.acquire)

    async def __aexit__(self, *exc):
        self._lk.release()


spawn = create_task



from dan.core.target import Target

class SourcesProvider(Target, internal=True):
    installed = False
    default = False

    async def available_versions(self):
        raise NotImplementedError(f'available_versions not implemented in {self.__class__.__name__}')



from logging import *
from termcolor import colored
from dan.core.terminal import TermLogHandler

# default level
getLogger().setLevel(INFO)


def merge(lhs, rhs):
    if type(lhs) != type(rhs):
        raise RuntimeError(f"cannot merge {type(lhs)} with {rhs}")
    if type(lhs) == dict:
        for k in rhs:
            if k in lhs:
                lhs[k] = merge(lhs[k], rhs[k])
            else:
                lhs[k] = rhs[k]
    elif type(lhs) == list:
        lhs.extend(rhs)
    return lhs


class bind_back:
    def __init__(self, fn, *args):
        self.fn = fn
        self.args = args

    def __call__(self, *args, **kwds):
        return self.fn(*args, *self.args, **kwds)


class ColoredFormatter(Formatter):
    COLORS = {
        "WARNING": bind_back(colored, "yellow"),
        "INFO": bind_back(colored, "green"),
        "DEBUG": bind_back(colored, "cyan"),
        "CRITICAL": bind_back(colored, "yellow"),
        "ERROR": bind_back(colored, "red"),
    }

    COLORS_ATTRS = {
        "WARNING": list(),
        "INFO": list(),
        "DEBUG": list(),
        "CRITICAL": ["blink"],
        "ERROR": ["blink"],
    }

    COLOR_FORMAT = (
        f"[{colored('%(asctime)s.%(msecs)03d', 'grey')}]"
        f"[%(levelname)s] {colored('%(name)s:', 'white', attrs=['bold'])} %(message)s "
    )  # \
    # f"({colored('%(filename)s:%(lineno)d', 'grey')})"

    FORMAT = "[%(asctime)s.%(msecs)03d]" "[%(levelname)s] %(name)s: %(message)s "  # \
    # "(%(filename)s:%(lineno)d)"

    def __init__(self, use_color=True):
        super().__init__(
            self.COLOR_FORMAT if use_color else self.FORMAT, datefmt="%H:%M:%S"
        )
        self.use_color = use_color

    def format(self, record):
        levelname = record.levelname
        if self.use_color and levelname in self.COLORS:
            record.levelname = self.COLORS[levelname](
                levelname, attrs=["bold", *self.COLORS_ATTRS[levelname]]
            )
            record.msg = self.COLORS[levelname](record.msg, attrs=[])
        return super().format(record)


_color_formatter = ColoredFormatter()
_no_color_formatter = ColoredFormatter(use_color=False)


def setup_logger(logger: Logger):
    if not logger.hasHandlers():
        console = TermLogHandler()
        console.setFormatter(_color_formatter)
        logger.addHandler(console)
    else:
        for handler in logger.handlers:
            handler.setFormatter(
                _color_formatter
                if isinstance(handler, StreamHandler)
                else _no_color_formatter
            )


TRACE = DEBUG - 5
addLevelName(TRACE, "TRACE")

STATUS = INFO - 1
addLevelName(STATUS, "STATUS")

PROGRESS = INFO + 1
addLevelName(PROGRESS, "PROGRESS")


class ColoredLogger(Logger):
    def __init__(self, name):
        super().__init__(name)
        self.propagate = False
        setup_logger(self)

    def trace(self, message, *args, **kwargs):
        if self.isEnabledFor(TRACE):
            self._log(TRACE, message, args, **kwargs)


setLoggerClass(ColoredLogger)


class Logging:
    def get_logger(self, name=None):
        logger = getattr(self, "_logger", None)
        if logger is None:
            if name is None:
                name = getattr(self, "fullname", self.__class__.__name__)
            if name.startswith("root."):
                name = name.removeprefix("root.")
            logger = getLogger(name)
            setattr(self, "_logger", logger)
        return logger

    def trace(self, msg, *args, **kwargs):
        self.get_logger().trace(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.get_logger().debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.get_logger().info(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.get_logger().warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.get_logger().error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.get_logger().critical(msg, *args, **kwargs)


def _get_makefile_logger():
    from dan.core.include import context

    if not hasattr(context.current, "_logger"):
        makefile_logger = getLogger(context.current.name)
        setup_logger(makefile_logger)
        setattr(context.current, "_logger", makefile_logger)
    else:
        makefile_logger: Logger = getattr(context.current, "_logger")
    return makefile_logger


def trace(*args, **kwds):
    return _get_makefile_logger().trace(*args, **kwds)


def debug(*args, **kwds):
    return _get_makefile_logger().debug(*args, **kwds)


def info(*args, **kwds):
    return _get_makefile_logger().info(*args, **kwds)


def warning(*args, **kwds):
    return _get_makefile_logger().warning(*args, **kwds)


def error(*args, **kwds):
    return _get_makefile_logger().error(*args, **kwds)


def critical(*args, **kwds):
    return _get_makefile_logger().critical(*args, **kwds)


class lazy_fmt:
    def __init__(self, fn):
        self.__fn = fn

    def __str__(self):
        return self.__fn()



from dataclasses import dataclass, field
import functools
import itertools
import os
import fnmatch

from dataclasses_json import dataclass_json
import sys
from collections.abc import Iterable

from dan import logging

import dan.core.typing as t
from dan.core import diagnostics as diag
from dan.core.cache import Cache
from dan.core.makefile import MakeFile
from dan.core.pathlib import Path
from dan.core.include import MakeFileError, include_makefile, Context
from dan.core import aiofiles, asyncio
from dan.core.requirements import RequiredPackage, load_requirements
from dan.core.settings import InstallMode, InstallSettings, Settings
from dan.core.test import Test
from dan.core.utils import unique
from dan.cxx import init_toolchains
from dan.core.target import Option, Target
from dan.cxx.targets import Executable
from dan.core.runners import max_jobs
from dan.core.terminal import TerminalMode, TermStream, set_mode as set_terminal_mode


def flatten(list_of_lists):
    if len(list_of_lists) == 0:
        return list_of_lists
    if isinstance(list_of_lists[0], Iterable):
        return flatten(list_of_lists[0]) + flatten(list_of_lists[1:])
    return list_of_lists[:1] + flatten(list_of_lists[1:])


@dataclass_json
@dataclass
class Config:
    source_path: Path = None
    build_path: Path = None
    toolchain: str = None
    settings: Settings = field(default_factory=lambda: Settings())


class ConfigCache(Cache[Config]):
    indent = 4


def _get_code_position(code, instruction_index):
    if instruction_index < 0 or not hasattr(
        code, "co_positions"
    ):  # Python <= 3.10 does not have co_position
        return (None, 0, 0, 0)
    positions_gen = code.co_positions()
    return next(itertools.islice(positions_gen, instruction_index // 2, None))


def _tb_entries(tb):
    entries = list()
    while tb is not None:
        entries.append(tb)
        tb = tb.tb_next
    return entries


def _walk_tb(tb, reverse=True):
    if isinstance(tb, Exception):
        tb = tb.__traceback__

    entries = _tb_entries(tb)

    if reverse:
        entries = reversed(entries)

    for tb in entries:
        positions = _get_code_position(tb.tb_frame.f_code, tb.tb_lasti)
        # Yield tb_lineno when co_positions does not have a line number to
        # maintain behavior with walk_tb.
        if positions[0] is None:
            yield tb.tb_frame.f_code.co_filename, (tb.tb_lineno,) + positions[1:]
        else:
            yield tb.tb_frame.f_code.co_filename, positions


def gen_python_diags(err: Exception):
    diagnostics = diag.DiagnosticCollection()
    if not diag.enabled:
        return diagnostics
    match err:
        case MakeFileError():
            cause = err.__cause__
            related = None
            if isinstance(cause, MakeFileError):
                related = list()
                cause_diags = gen_python_diags(cause)
                for filename, diagns in cause_diags.items():
                    for d in diagns:
                        related.append(
                            diag.RelatedInformation(
                                diag.Location(diag.Uri(filename), d.range), d.message
                            )
                        )
                diagnostics.update(cause_diags)
                err_filename = str(err.path)
                for filename, (lineno, end_lineno, colno, end_colno) in _walk_tb(cause):
                    if filename == err_filename:
                        break
            else:
                for filename, (lineno, end_lineno, colno, end_colno) in _walk_tb(cause):
                    break
            diagnostics[filename] = diag.Diagnostic(
                message=str(cause),
                range=diag.Range(
                    start=diag.Position(lineno - 1, colno),
                    end=diag.Position(end_lineno - 1, end_colno),
                ),
                code=cause.__class__.__name__,
                source="dan.make",
                related_information=related,
            )
        case _:
            for filename, (lineno, end_lineno, colno, end_colno) in _walk_tb(err):
                break
            diagnostics[filename] = diag.Diagnostic(
                message=str(err),
                range=diag.Range(
                    start=diag.Position(lineno - 1, colno),
                    end=diag.Position(end_lineno - 1, end_colno),
                ),
                code=err.__class__.__name__,
                source="dan.make",
            )

    return diagnostics


class Make(logging.Logging):
    _config_name = "dan.config.json"
    _cache_name = "dan.cache"

    def __init__(
        self,
        build_path: str,
        targets: list[str] = None,
        verbose: int = 0,
        for_install: bool = False,
        jobs: int = None,
        terminal_mode: TerminalMode = None,
        diags=False,
    ):
        jobs = jobs or os.cpu_count()
        max_jobs(jobs)

        match verbose:
            case 1:
                log_level = logging.DEBUG
            case 2:
                log_level = logging.TRACE
            case -1:
                log_level = logging.ERROR
            case 0:
                log_level = logging.INFO
            case _:
                logging.getLogger().warning(
                    "unknown verbosity level: %s, using INFO", verbose
                )
                log_level = logging.INFO

        logging.getLogger().setLevel(log_level)

        if terminal_mode is not None:
            set_terminal_mode(terminal_mode)

        self.term = TermStream("make")

        if diags:
            diag.enabled = True

        self.for_install = for_install

        self.build_path = Path(build_path)
        self.config_path = build_path / self._config_name
        self.cache_path = build_path / self._cache_name

        self.required_targets = targets
        sys.pycache_prefix = str(build_path / "__pycache__")
        self._config = ConfigCache.instance(self.config_path)
        self.cache = Cache.instance(self.cache_path, binary=True)

        self.debug(f"jobs: {jobs}")

        self._diagnostics = diag.DiagnosticCollection()

        self.context = Context()

    @property
    def config(self) -> Config:
        return self._config.data

    @property
    def settings(self) -> Settings:
        return self._config.data.settings

    @property
    def source_path(self):
        return Path(self.config.source_path)

    @property
    def toolchain(self):
        return self.context.get("cxx_target_toolchain")

    @property
    def root(self) -> MakeFile:
        return self.context.root

    @property
    def env(self) -> dict[str, str]:
        from dan.cxx.detect import get_dan_path

        env = self.toolchain.env
        epath = env.get("PATH", os.environ["PATH"]).split(os.pathsep)
        epath.insert(0, str(get_dan_path() / "os-utils" / "bin"))
        for t in self.executable_targets:
            parts = t.env.get("PATH", os.environ["PATH"]).split(os.pathsep)
            for p in parts:
                if not p in epath:
                    epath.append(p)
        env["PATH"] = os.pathsep.join(epath)
        return env

    async def configure(self, source_path: str, toolchain: str = None):
        self.config.source_path = str(source_path)
        self.config.build_path = str(self.build_path)
        self.info(f"source path: {self.config.source_path}")
        self.info(f"build path: {self.config.build_path}")
        if toolchain:
            self.config.toolchain = toolchain
        if not self.config.toolchain:
            self.warning("no toolchain configured")
        await self._config.save()

    @asyncio.cached
    async def initialize(self):
        assert self.config_path.exists(), "configure first"

        self.debug(f"source path: {self.source_path}")
        self.debug(f"build path: {self.build_path}")

        toolchain = self.config.toolchain
        build_type = self.settings.build_type

        self.info(f"using '{toolchain}' toolchain in '{build_type.name}' mode")

        with self.context:
            init_toolchains(toolchain, self.settings)
            try:
                include_makefile(self.source_path, self.build_path)
            except MakeFileError as err:
                self._diagnostics.update(gen_python_diags(err))
                raise

        target_toolchain = self.context.get("cxx_target_toolchain")
        target_toolchain.build_type = build_type
        if self.for_install:
            library_dest = (
                Path(self.settings.install.destination)
                / self.settings.install.libraries_prefix
            )
            target_toolchain.rpath = str(library_dest.absolute())

        self.debug(f"targets: {[t.name for t in self.targets]}")

    def __matches(self, target: Target | Test):
        for required in self.required_targets:
            if fnmatch.fnmatch(target.fullname, f"*{required}*"):
                return True
        return False

    @functools.cached_property
    def targets(self) -> list[Target]:
        items = list()
        if self.required_targets and len(self.required_targets) > 0:
            for target in self.root.all_targets:
                if self.__matches(target):
                    items.append(target)
        else:
            items = self.root.all_default
        return items

    @functools.cached_property
    def tests(self) -> list[Test]:
        items = list()
        if self.required_targets and len(self.required_targets) > 0:
            for required in self.required_targets:
                pos = required.find(":")
                if pos > 0:
                    test_name = required[:pos]
                    test_case = required[pos + 1 :]
                else:
                    test_name = required
                    test_case = None

                for test in self.root.all_tests:
                    if fnmatch.fnmatch(test.fullname, f"*{test_name}*"):
                        if len(test) > 1 and test_case is not None:
                            cases = list()
                            for case in test.cases:
                                if fnmatch.fnmatch(case.name, test_case):
                                    cases.append(case)
                            if len(cases) == 0:
                                self.warning(
                                    "couldn't find any test case in %s matching '%s'",
                                    test.name,
                                    test_case,
                                )
                            test.cases = cases

                        items.append(test)
        else:
            for test in self.root.all_tests:
                items.append(test)
        return items

    @property
    def all_options(self) -> list[Option]:
        opts = []
        for target in self.targets:
            for o in target.options:
                opts.append(o)
        for makefile in self.context.all_makefiles:
            for o in makefile.options:
                opts.append(o)
        return opts

    @property
    def diagnostics(self):
        result: diag.DiagnosticCollection = diag.DiagnosticCollection()
        if self.root:
            for target in self.root.all_targets:
                result.update(target.diagnostics)
        result.update(self._diagnostics)
        return result

    async def targets_of(self, sources: list[Path]) -> dict[Path, Target | None]:
        from dan.cxx.targets import CXXObjectsTarget

        result = {Path(source): None for source in sources}

        def check(t: CXXObjectsTarget):
            nonlocal result
            for source in result.keys():
                if t.source_path not in source.parents:
                    continue
                if source.suffix[1].lower() == "h":
                    for p in [*t.includes.private_raw, *t.includes.public_raw]:
                        p: Path = p
                        if p not in source.parents:
                            continue
                        result[source] = t
                else:
                    t._init_sources()
                    if source.name in [Path(s).name for s in t.sources]:
                        result[source] = t

        object_targets = [
            target
            for target in self.root.all_targets
            if isinstance(target, CXXObjectsTarget)
        ]

        def _on_done(task: asyncio.Task):
            nonlocal result
            if not task.cancelled():
                done = None not in result.values()
                if done:
                    g.cancel()

        async with asyncio.TaskGroup() as g:
            for t in object_targets:
                g.create_task(asyncio.async_wait(check, t)).add_done_callback(_on_done)

        return result

    async def apply_options(self, *options):
        await self.initialize()
        from dan.core.settings import _apply_inputs

        all_opts = self.all_options

        def get_option(name):
            for opt in all_opts:
                if opt.fullname == name:
                    return opt.cache, opt.value, opt.type

        _apply_inputs(options, get_option, logger=self, input_type_name="option")

    async def apply_settings(self, *settings):
        from dan.core.settings import apply_settings

        apply_settings(self.settings, *settings, logger=self)

    @staticmethod
    def toolchains():
        from dan.cxx.detect import get_toolchains

        return get_toolchains()

    async def _build_target(self, t: Target):
        try:
            await t.build()
        except Exception as err:
            self._diagnostics.update(gen_python_diags(err))
            raise

    async def get_all_dependencies(self, target, deps: set):
        if target in deps:
            return
        await load_requirements(
            target.requires, makefile=self.root, logger=self, install=True
        )
        for d in target.dependencies.all:
            match d:
                case RequiredPackage():
                    if d.target:
                        deps.add(d.target)
                        await self.get_all_dependencies(d.target, deps)
                case Target():
                    deps.add(d)
                    await self.get_all_dependencies(d, deps)

    async def build(self, targets: list[Target] = None):
        await self.initialize()

        if targets is None:
            targets = self.targets

        all_targets = set()
        async with self.term.task_group("installing dependencies...") as g:
            for t in targets:
                g.create_task(self.get_all_dependencies(t, all_targets))

        all_targets.update(targets)

        self.term.status("building...")
        async with self.term.task_group("building...") as g:
            for t in all_targets:
                g.create_task(self._build_target(t))

        self.term.status("done", icon="✔")

    async def _install_target_deps(self, t: Target):
        deps_install_path = self.root.pkgs_path
        deps_settings = InstallSettings(deps_install_path)
        try:
            await load_requirements(
                t.requires, makefile=self.root, logger=self, install=True
            )

        except Exception as err:
            self._diagnostics.update(gen_python_diags(err))
            raise

    async def install_dependencies(self, targets: list[Target] = None):
        await self.initialize()

        if targets is None:
            targets = self.targets

        async with self.term.task_group("installing dependencies...") as g:
            for target in targets:
                g.create_task(self._install_target_deps(target))

    async def _install_target(self, t: Target, mode: InstallMode):
        try:
            return await t.install(self.settings.install, mode)
        except Exception as err:
            self._diagnostics.update(gen_python_diags(err))
            raise

    async def install(self, mode: InstallMode = InstallMode.user):
        await self.initialize()

        self.for_install = True
        await self.initialize()
        targets = []
        for t in self.targets:
            if t.installed:
                targets.append(t)
            else:
                self.debug("skipping %s (not marked as installed)", t.fullname)

        await self.build(targets)

        async with self.term.task_group("installing targets...") as g:
            for target in targets:
                g.create_task(self._install_target(target, mode))

        installed_files = unique(flatten(g.results()))
        manifest_path = (
            self.settings.install.data_destination
            / "dan"
            / f"{self.root.name}-manifest.txt"
        )
        manifest_path.parent.mkdir(parents=True, exist_ok=True)

        async with aiofiles.open(manifest_path, "w") as f:
            await f.writelines(
                [
                    os.path.relpath(p, manifest_path.parent) + "\n"
                    for p in installed_files
                ]
            )

    async def package(self, pkg_type: str, mode: InstallMode = InstallMode.user):
        import tempfile

        targets = self.targets
        if len(targets) == 1:
            base_name = targets[0].name
            version = targets[0].version
        else:
            base_name = self.root.name

        if not version:
            version = self.root.version

        if version:
            base_name += f"-v{version}"

        archive_name = f"{base_name}.{pkg_type}"

        with tempfile.TemporaryDirectory(prefix=archive_name) as tmp_dir:
            self.settings.install.destination = tmp_dir
            tmp_dir = Path(tmp_dir)
            if pkg_type.startswith("tar"):
                import tarfile

                comp_type = pkg_type.split(".")[1]
                self.info("creating %s", archive_name)
                await self.install(mode)
                with tarfile.open(archive_name, f"w:{comp_type}") as archive:
                    for it in tmp_dir.iterdir():
                        archive.add(it, arcname=it.relative_to(tmp_dir))
            elif pkg_type == "zip":
                import zipfile

                self.info("creating %s", archive_name)
                await self.install(mode)
                with zipfile.ZipFile(archive_name, "w") as archive:
                    for folder_name, sub_folders, file_names in os.walk(tmp_dir):
                        for filename in file_names:
                            file_path = Path(folder_name) / filename
                            archive.write(file_path, file_path.relative_to(tmp_dir))
            else:
                raise RuntimeError(f"Unhandled package type: {pkg_type}")

    @property
    def executable_targets(self) -> list[Executable]:
        return [exe for exe in self.targets if isinstance(exe, Executable)]

    async def scan_toolchains(self, script: Path = None):
        from dan.cxx.detect import create_toolchains, load_env_toolchain

        if script:
            load_env_toolchain(script)
        else:
            create_toolchains()

    async def run(self):
        await self.initialize()
        with self.context:
            results = await asyncio.gather(
                *[t.execute(log=True) for t in self.executable_targets]
            )
            for result in results:
                if result[2] != 0:
                    return result[2]
        return 0

    async def _test_target(self, t: Test):
        try:
            return await t.run_test()
        except Exception as err:
            self._diagnostics.update(gen_python_diags(err))
            raise

    async def test(self):
        await self.initialize()
        tests = self.tests
        if len(tests) == 0:
            self.error("No test selected")
            return 255

        with self.context:
            async with self.term.task_group("testing...") as g:
                for test in tests:
                    g.create_task(self._test_target(test))

            if all(g.results()):
                self.info("Success !")
                return 0
            else:
                self.error("Failed !")
                return 255

    async def clean(self):
        await self.initialize()
        with self.context:
            await asyncio.gather(*[t.clean() for t in self.targets])
            # from dan.cxx import target_toolchain
            # target_toolchain.compile_commands.clear()
